@@ -17,4 +17,10 @@
 # along with this program.  If not, see <http://www.gnu.org/licenses/>.
 #
 
-__version__ = '0.7.3'
+__all__ = ['__version__', ]
+__version__ = '0.8.0'
+
+
+class Error(Exception):
+    """Base class for all package errors"""
+    pass
